# SPDX-License-Identifier: AGPL-3.0-or-later OR GPL-2.0-or-later OR CERN-OHL-S-2.0+ OR Apache-2.0
# type: ignore
import unittest

from pdkmaster import _util
from pdkmaster.technology import geometry as _geo
from pdkmaster.design import layout as _lay
from pdkmaster.design.layout import layout_ as _laylay
from pdkmaster.io import klayout as _kl
from pdkmaster.io.klayout import merge_ as _klmrg

from ...dummy import MyNet, dummy_tech, dummy_cktfab, dummy_layoutfab
prims = dummy_tech.primitives


class SupportTest(unittest.TestCase):
    def test_importexport(self):
        import pya

        rect = _geo.Rect.from_floats(values=(0.0, 0.0, 1.0, 1.0))
        polygon = _geo.Polygon.from_floats(points=(
            (0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 2.0),
            (2.0, 2.0), (2.0, 0.0), (0.0, 0.0),
        ))
        eps = _geo.epsilon
        box = pya.Box(0, 0, round(1.0/eps), round(1.0/eps))

        self.assertEqual(_klmrg._import_regionshape(_klmrg._export_polygon(rect)), rect)
        self.assertEqual(_klmrg._import_polygon(box), rect)
        self.assertEqual(_klmrg._import_regionshape(_klmrg._export_polygon(polygon)), polygon)


class MPSDictTest(unittest.TestCase):
    def test_in(self):
        rect1 = _geo.Rect(left=0.0, bottom=0.0, right=1.0, top=1.0)
        rect2 = _geo.Rect(left=1.0, bottom=0.0, right=2.0, top=1.0)
        rect3 = _geo.Rect(left=2.0, bottom=0.0, right=3.0, top=1.0)
        rect12 = _geo.Rect(left=0.0, bottom=0.0, right=2.0, top=1.0)
        rect123 = _geo.Rect(left=0.0, bottom=0.0, right=3.0, top=1.0)

        mps12 = _geo.MultiPartShape(fullshape=rect12, parts=(rect1, rect2))
        mps123 = _geo.MultiPartShape(fullshape=rect123, parts=(rect1, rect2, rect3))

        d = _klmrg._MPSDict(merger=_klmrg._ShapeMerger())
        e = d[mps12]

        self.assertIsInstance(e, _klmrg._MPSDictElem)
        self.assertEqual(e.mps_orig, mps12)
        self.assertEqual(len(e.partregions), 2)
        self.assertTrue(mps12 in d)
        self.assertFalse(mps123 in d)

    def test_merge_with(self):
        rect1 = _geo.Rect(left=0.0, bottom=0.0, right=1.0, top=1.0)
        rect2 = _geo.Rect(left=1.0, bottom=0.0, right=2.0, top=1.0)
        rect3 = _geo.Rect(left=2.0, bottom=0.0, right=3.0, top=1.0)
        rect4 = _geo.Rect(left=3.0, bottom=0.0, right=4.0, top=1.0)
        rect12 = _geo.Rect(left=0.0, bottom=0.0, right=2.0, top=1.0)
        rect34 = _geo.Rect(left=2.0, bottom=0.0, right=4.0, top=1.0)
        rect1234 = _geo.Rect(left=0.0, bottom=0.0, right=4.0, top=1.0)

        mps12 = _geo.MultiPartShape(fullshape=rect12, parts=(rect1, rect2))
        mps34 = _geo.MultiPartShape(fullshape=rect34, parts=(rect3, rect4))
        mps1234 = _geo.MultiPartShape(fullshape=rect1234, parts=(rect1, rect2, rect3, rect4))

        d =_klmrg._MPSDict()

        elem1 = d[mps12]
        elem2 = d[mps34]

        ref1 = _klmrg._PartRef(elem=elem1, idx=0)
        ref2 = _klmrg._PartRef(elem=elem1, idx=1)
        ref3 = _klmrg._PartRef(elem=elem2, idx=0)
        ref4 = _klmrg._PartRef(elem=elem2, idx=1)

        off1 = elem1.merge_with(elem1)
        off2 = elem1.merge_with(elem2)

        part1 = ref1.deref()
        part2 = ref2.deref()
        part3 = ref3.deref()
        part4 = ref4.deref()

        self.assertEqual(off1, 0)
        self.assertEqual(off2, 2)
        self.assertIsInstance(part1, _geo.MultiPartShape._Part)
        self.assertEqual(part1.partshape, rect1)
        self.assertEqual(part1.multipartshape, mps1234)
        self.assertIsInstance(part2, _geo.MultiPartShape._Part)
        self.assertEqual(part2.partshape, rect2)
        self.assertEqual(part2.multipartshape, mps1234)
        self.assertIsInstance(part3, _geo.MultiPartShape._Part)
        self.assertEqual(part3.partshape, rect3)
        self.assertEqual(part3.multipartshape, mps1234)
        self.assertIsInstance(part4, _geo.MultiPartShape._Part)
        self.assertEqual(part4.partshape, rect4)
        self.assertEqual(part4.multipartshape, mps1234)


class ShapeMergerTest(unittest.TestCase):
    def test_part_overlap(self):
        rect1 = _geo.Rect(left=0.0, bottom=0.0, right=1.0, top=1.0)
        rect2 = _geo.Rect(left=0.5, bottom=0.0, right=2.0, top=1.0)
        rect3 = _geo.Rect(left=2.0, bottom=0.0, right=3.0, top=1.0)
        rect12 = _geo.Rect(left=0.0, bottom=0.0, right=2.0, top=1.0)
        rect13 = _geo.Rect(left=0.0, bottom=0.0, right=3.0, top=1.0)
        mps1 = _geo.MultiPartShape(fullshape=rect12, parts=(rect1, rect2))
        mps2 = _geo.MultiPartShape(fullshape=rect13, parts=(rect1, rect3))

        # Overlapping parts
        with self.assertRaises(ValueError):
            with _klmrg._ShapeMerger() as merger:
                merger.mps_lookup[mps1]
        # Mismatch fullshape and parts
        with self.assertRaises(ValueError):
            with _klmrg._ShapeMerger() as merger:
                merger.mps_lookup[mps2]

    def test_part_poly(self):
        rect1 = _geo.Rect(left=0.0, bottom=0.0, right=1.0, top=1.0)
        rect2 = _geo.Rect(left=1.0, bottom=0.0, right=2.0, top=1.0)
        rect3 = _geo.Rect(left=2.0, bottom=0.0, right=3.0, top=1.0)
        rect12 = _geo.Rect(left=0.0, bottom=0.0, right=2.0, top=1.0)
        rect23 = _geo.Rect(left=1.0, bottom=0.0, right=3.0, top=1.0)
        rect123 =  _geo.Rect(left=0.0, bottom=0.0, right=3.0, top=1.0)

        mps12 = _geo.MultiPartShape(fullshape=rect12, parts=(rect1, rect2))
        mps123 = _geo.MultiPartShape(fullshape=rect123, parts=(rect1, rect23))

        merged = _klmrg._ShapeMerger()((
            mps12.parts[0],
            _geo.MultiShape(shapes=(mps12.parts[1], rect3)),
        ))

        self.assertEqual(len(merged), 2)

        part0 = merged[0].deref()
        part1 = merged[1].deref()

        self.assertIsInstance(part0, _geo.MultiPartShape._Part)
        self.assertIsInstance(part1, _geo.MultiPartShape._Part)

        self.assertEqual(part0.partshape, rect1)
        self.assertEqual(part1.partshape, rect23)
        self.assertEqual(part0.multipartshape, mps123)
        self.assertEqual(part0.multipartshape, part1.multipartshape)

    def test_parts(self):
        rect1 = _geo.Rect(left=0.0, bottom=0.0, right=1.0, top=1.0)
        rect2 = _geo.Rect(left=1.0, bottom=0.0, right=2.0, top=1.0)
        rect3 = _geo.Rect(left=2.0, bottom=0.0, right=3.0, top=1.0)
        rect12 = _geo.Rect(left=0.0, bottom=0.0, right=2.0, top=1.0)
        rect23 = _geo.Rect(left=1.0, bottom=0.0, right=3.0, top=1.0)
        rect123 = _geo.Rect(left=0.0, bottom=0.0, right=3.0, top=1.0)

        mps1 = _geo.MultiPartShape(fullshape=rect12, parts=(rect1, rect2))
        mps2 = _geo.MultiPartShape(fullshape=rect23, parts=(rect2, rect3))
        mps3 = _geo.MultiPartShape(fullshape=rect123, parts=(rect1, rect2, rect3))

        merged = _klmrg._ShapeMerger()((
            mps1.parts[0],
            _geo.MultiShape(shapes=(mps1.parts[1], mps2.parts[0])),
            mps2.parts[1],
        ))

        part0 = merged[0].deref()
        part1 = merged[1].deref()
        part2 = merged[2].deref()

        self.assertIsInstance(part0, _geo.MultiPartShape._Part)
        self.assertIsInstance(part1, _geo.MultiPartShape._Part)
        self.assertIsInstance(part2, _geo.MultiPartShape._Part)
        self.assertEqual(part0.partshape, rect1)
        self.assertEqual(part1.partshape, rect2)
        self.assertEqual(part2.partshape, rect3)
        self.assertEqual(part0.multipartshape, mps3)
        self.assertEqual(part0.multipartshape, part1.multipartshape)
        self.assertEqual(part0.multipartshape, part2.multipartshape)

    def test_parts_2(self):
        rect1 = _geo.Rect(left=0.0, bottom=0.0, right=1.0, top=1.0)
        rect2 = _geo.Rect(left=1.0, bottom=0.0, right=2.0, top=1.0)
        rect3 = _geo.Rect(left=2.0, bottom=0.0, right=3.0, top=1.0)
        rect4 = _geo.Rect(left=3.0, bottom=0.0, right=4.0, top=1.0)
        rect5 = _geo.Rect(left=4.0, bottom=0.0, right=5.0, top=1.0)
        rect6 = _geo.Rect(left=5.0, bottom=0.0, right=6.0, top=1.0)
        rect7 = _geo.Rect(left=6.0, bottom=0.0, right=7.0, top=1.0)
        rect8 = _geo.Rect(left=7.0, bottom=0.0, right=8.0, top=1.0)

        rect12 = _geo.Rect(left=0.0, bottom=0.0, right=2.0, top=1.0)
        rect23 = _geo.Rect(left=1.0, bottom=0.0, right=3.0, top=1.0)
        rect34 = _geo.Rect(left=2.0, bottom=0.0, right=4.0, top=1.0)
        rect45 = _geo.Rect(left=3.0, bottom=0.0, right=5.0, top=1.0)
        rect56 = _geo.Rect(left=4.0, bottom=0.0, right=6.0, top=1.0)
        rect67 = _geo.Rect(left=5.0, bottom=0.0, right=7.0, top=1.0)
        rect78 = _geo.Rect(left=6.0, bottom=0.0, right=8.0, top=1.0)
        rect1_8 = _geo.Rect(left=0.0, bottom=0.0, right=8.0, top=1.0)

        mps23 = _geo.MultiPartShape(fullshape=rect23, parts=(rect2, rect3))
        mps45 = _geo.MultiPartShape(fullshape=rect45, parts=(rect4, rect5))
        mps67 = _geo.MultiPartShape(fullshape=rect67, parts=(rect6, rect7))
        mps1_8 = _geo.MultiPartShape(
            fullshape=rect1_8, parts=(rect12, rect34, rect56, rect78)
        )

        merged = _klmrg._ShapeMerger()((
            _geo.MultiShape(shapes=(rect1, mps23.parts[0])),
            _geo.MultiShape(shapes=(mps23.parts[1], mps45.parts[0])),
            _geo.MultiShape(shapes=(mps45.parts[1], mps67.parts[0])),
            _geo.MultiShape(shapes=(mps67.parts[1], rect8)),
        ))

        for i, partref in enumerate(merged):
            part = partref.deref()
            msg = f"part{i}"
            self.assertIsInstance(part, _geo.MultiPartShape._Part, msg=msg)
            self.assertEqual(part.multipartshape, mps1_8, msg=msg)
            self.assertEqual(part.partshape, mps1_8.parts[i].partshape, msg=msg)

    def test_part_rect(self):
        rect1 = _geo.Rect(left=0.0, bottom=0.0, right=1.0, top=1.0)
        rect2 = _geo.Rect(left=1.0, bottom=0.0, right=2.0, top=1.0)
        rect3 = _geo.Rect(left=2.0, bottom=0.0, right=3.0, top=1.0)
        rect4 = _geo.Rect(left=2.5, bottom=0.0, right=4.0, top=1.0)
        rect12 = _geo.Rect(left=0.0, bottom=0.0, right=2.0, top=1.0)
        rect23 = _geo.Rect(left=1.0, bottom=0.0, right=3.0, top=1.0)
        rect34 = _geo.Rect(left=2.0, bottom=0.0, right=4.0, top=1.0)
        rect1234 = _geo.Rect(left=1.0, bottom=0.0, right=3.0, top=1.0)

        mps1 = _geo.MultiPartShape(fullshape=rect23, parts=(rect2, rect3))
        mps2 = _geo.MultiPartShape(fullshape=rect1234, parts=(rect12, rect34))

        merged = _klmrg._ShapeMerger()((
            _geo.MultiShape(shapes=(rect1, mps1.parts[0])),
            _geo.MultiShape(shapes=(mps1.parts[1], rect4)),
        ))

        part0 = merged[0].deref()
        part1 = merged[1].deref()

        self.assertIsInstance(part0, _geo.MultiPartShape._Part)
        self.assertIsInstance(part1, _geo.MultiPartShape._Part)
        self.assertEqual(part0.multipartshape, mps2)
        self.assertEqual(part0.multipartshape, part1.multipartshape)

    def test_part_rect_2(self):
        rect1_1 = _geo.Rect(left=0.0, bottom=0.0, right=1.0, top=1.0)
        rect1_2 = _geo.Rect(left=1.0, bottom=0.0, right=2.0, top=1.0)
        rect1_3 = _geo.Rect(left=2.0, bottom=0.0, right=3.0, top=1.0)
        rect1_123 = _geo.Rect(left=0.0, bottom=0.0, right=3.0, top=1.0)

        rect2_1 = _geo.Rect(left=0.0, bottom=2.0, right=1.0, top=3.0)
        rect2_2 = _geo.Rect(left=1.0, bottom=2.0, right=2.0, top=3.0)
        rect2_3 = _geo.Rect(left=2.0, bottom=2.0, right=3.0, top=3.0)
        rect2_123 = _geo.Rect(left=0.0, bottom=2.0, right=3.0, top=3.0)

        rect12_1 = _geo.Rect(left=-1, bottom=0.0, right=0.5, top=3.0)
        rect12_2 = _geo.Rect(left=2.5, bottom=0.0, right=4, top=3.0)

        poly12_1 = _geo.Polygon.from_floats(points=(
            (-1.0, 0.0), (-1.0, 3.0), (1.0, 3.0), (1.0, 2.0),
            (0.5, 2.0), (0.5, 1.0), (1.0, 1.0), (1.0, 0.0), (-1.0, 0.0),
        ))
        poly12_2 = _geo.Polygon.from_floats(points=(
            (2.0, 0.0), (2.0, 1.0), (2.5, 1.0), (2.5, 2.0), (2.0, 2.0), (2.0, 3.0),
            (4.0, 3.0), (4.0, 0.0), (2.0, 0.0),
        ))

        poly12_12 = _geo.Polygon.from_floats(points=(
            (-1.0, 0.0), (-1.0, 2.0), (0.5, 2.0), (0.5, 1.0), (2.5, 1.0), (2.5, 2.0),
            (-1.0, 2.0), (-1.0, 3.0), (4.0, 3.0), (4.0, 0.0), (-1.0, 0.0),
        ))

        mps1 = _geo.MultiPartShape(fullshape=rect1_123, parts=(rect1_1, rect1_2, rect1_3))
        mps2 = _geo.MultiPartShape(fullshape=rect2_123, parts=(rect2_1, rect2_2, rect2_3))
        mps12 = _geo.MultiPartShape(fullshape=poly12_12, parts=(
            poly12_1, rect1_2, rect2_2, poly12_2
        ))

        shapes = (
            _geo.MultiShape(shapes=(rect12_1, mps1.parts[0], mps2.parts[0])),
            _geo.MultiShape(shapes=(mps1.parts[1], mps2.parts[1])),
            _geo.MultiShape(shapes=(rect12_2, mps1.parts[2], mps2.parts[2])),
        )

        (part0ref, ms1, part2ref) = _klmrg._ShapeMerger()(shapes)
        part0 = part0ref.deref()
        part2 = part2ref.deref()
        (part1_0, part1_1) = ms1.deref().shapes

        self.assertIsInstance(part0, _geo.MultiPartShape._Part)
        self.assertIsInstance(part1_0, _geo.MultiPartShape._Part)
        self.assertIsInstance(part1_1, _geo.MultiPartShape._Part)
        self.assertIsInstance(part2, _geo.MultiPartShape._Part)
        self.assertEqual(part0.multipartshape, mps12)
        self.assertEqual(part0.multipartshape, part1_0.multipartshape)
        self.assertEqual(part0.multipartshape, part1_1.multipartshape)
        self.assertEqual(part0.multipartshape, part2.multipartshape)


class MergeTest(unittest.TestCase):
    def test_type(self):
        with self.assertRaises(TypeError):
            _kl.merge(1)
        with self.assertRaises(TypeError):
            _kl.merge((1,))

    def test_mutable(self):
        # Mutable objects are not hashable
        rect1 = _geo.Rect(left=0.0, bottom=0.0, right=1.0, top=1.0)
        rect2 = _geo.Rect(left=1.0, bottom=0.0, right=2.0, top=1.0)
        rect12 = _geo.Rect(left=0.0, bottom=0.0, right=2.0, top=1.0)
        mps1 = _geo.MultiPartShape(fullshape=rect12, parts=(rect1, rect2))
        d = _klmrg._MPSDict()
        elem = _klmrg._MPSDictElem(mps=mps1, mpsdict=d)
        elem.init_elemmps()

        with self.assertRaises(TypeError):
            hash(d)
        with self.assertRaises(TypeError):
            hash(elem)
        with self.assertRaises(TypeError):
            hash(elem.elemmps)

    def test_zeroarea(self):
        p1 = _geo.Point(x=0.0, y=0.0)
        p2 = _geo.Point(x=1.0, y=1.0)
        l1 = _geo.Line(point1=p1, point2=p2)
        ms1 = _geo.MultiShape(shapes=(p1, l1))

        self.assertEqual(_kl.merge(p1), p1)

        merged = _kl.merge(ms1)
        self.assertIsInstance(merged, _geo.MultiShape)
        self.assertEqual(merged, ms1)

    def test_rect(self):
        rect1 = _geo.Rect(left=0.0, bottom=0.0, right=1.0, top=1.0)
        rect2 = _geo.Rect(left=0.0, bottom=1.0, right=1.0, top=2.0)
        rect3 = _geo.Rect(left=0.0, bottom=0.0, right=1.0, top=2.0)
        ms1 = _geo.MultiShape(shapes=(rect1, rect2))

        self.assertEqual(_kl.merge(rect1), rect1)
        self.assertEqual(_kl.merge(ms1), rect3)

    def test_rect_polygon(self):
        rect1 = _geo.Rect(left=0.0, bottom=0.0, right=1.0, top=1.0)
        polygon1 = _geo.Polygon.from_floats(points=(
            (1.0, 0.0), (1.0, 1.0), (2.0, 1.0), (2.0, 2.0),
            (3.0, 2.0), (3.0, 0.0), (1.0, 0.0),
        ))
        polygon2 = _geo.Polygon.from_floats(points=(
            (0.0, 0.0), (0.0, 1.0), (2.0, 1.0), (2.0, 2.0),
            (3.0, 2.0), (3.0, 0.0), (0.0, 0.0),
        ))

        ms = _geo.MultiShape(shapes=(rect1, polygon1))

        shape = _kl.merge(ms)

        self.assertEqual(shape, polygon2)

    def test_maskshape(self):
        mask = _util.get_first_of(dummy_tech.designmasks)

        rect1 = _geo.Rect(left=0.0, bottom=0.0, right=1.0, top=1.0)
        poly2 = _geo.Polygon.from_floats(points=(
            (1.0, 0.0), (1.0, 1.0), (2.0, 1.0), (2.0, 2.0), (3.0, 2.0), (3.0, 0.0),
            (1.0, 0.0),
        ))
        poly12 = _geo.Polygon.from_floats(points=(
            (0.0, 0.0), (0.0, 1.0), (2.0, 1.0), (2.0, 2.0), (3.0, 2.0), (3.0, 0.0),
            (0.0, 0.0),
        ))
        mps1 = _geo.MultiPartShape(fullshape=poly12, parts=(rect1, poly2))
        ms1 = _geo.MaskShape(mask=mask, shape=rect1)
        ms2 = _geo.MaskShape(mask=mask, shape=mps1.parts[0])

        self.assertEqual(_kl.merge(ms1), ms1)
        self.assertEqual(_kl.merge(ms2), ms2)

    def test_maskshapes(self):
        mask = _util.get_first_of(dummy_tech.designmasks)

        rect1 = _geo.Rect(left=0.0, bottom=0.0, right=1.0, top=1.0)
        rect2 = _geo.Rect(left=2.0, bottom=0.0, right=3.0, top=1.0)
        ms1 = _geo.MaskShape(mask=mask, shape=rect1)
        ms2 = _geo.MaskShape(mask=mask, shape=rect2)
        mss1 = _geo.MaskShapes((ms1, ms2))

        self.assertEqual(_kl.merge(mss1), mss1)

    def test_layout(self):
        metal = prims.metal
        fab = _lay.LayoutFactory(tech=dummy_tech)

        net1 = MyNet(name="net1")
        rect1 = _geo.Rect(left=0.0, bottom=0.0, right=1.0, top=1.0)
        layout1 = fab.new_layout()
        layout1.add_shape(layer=metal, net=net1, shape=rect1)

        layout1_merged = _kl.merge(layout1)
        self.assertEqual(layout1_merged, layout1)

    def test_layout_multishapeparts(self):
        active = prims.active

        r1 = _geo.Rect.from_floats(values=(0.0, 0.0, 1.0, 1.0))
        r2 = _geo.Rect.from_floats(values=(1.0, 0.0, 2.0, 1.0))
        r12 = _geo.Rect.from_floats(values=(0.0, 0.0, 2.0, 1.0))
        r3 = _geo.Rect.from_floats(values=(2.0, 0.0, 3.0, 1.0))
        r4 = _geo.Rect.from_floats(values=(3.0, 0.0, 4.0, 1.0))
        r234 = _geo.Rect.from_floats(values=(1.0, 0.0, 4.0, 1.0))
        r5 = _geo.Rect.from_floats(values=(4.0, 0.0, 5.0, 1.0))
        r45 = _geo.Rect.from_floats(values=(3.0, 0.0, 5.0, 1.0))
        r6 = _geo.Rect.from_floats(values=(5.0, 0.0, 6.0, 1.0))
        r7 = _geo.Rect.from_floats(values=(6.0, 0.0, 7.0, 1.0))
        r567 = _geo.Rect.from_floats(values=(4.0, 0.0, 7.0, 1.0))
        r8 = _geo.Rect.from_floats(values=(7.0, 0.0, 8.0, 1.0))
        r78 = _geo.Rect.from_floats(values=(6.0, 0.0, 8.0, 1.0))
        r1_8 = _geo.Rect.from_floats(values=(0.0, 0.0, 8.0, 1.0))

        mps234 = _geo.MultiPartShape(fullshape=r234, parts=(r2, r3, r4))
        mps567 = _geo.MultiPartShape(fullshape=r567, parts=(r5, r6, r7))

        ckt = dummy_cktfab.new_circuit(name="ckt")
        layouter = dummy_layoutfab.new_circuitlayouter(circuit=ckt, boundary=None)
        layout = layouter.layout

        net1 = ckt.new_net(name="net1", external=False)
        net2 = ckt.new_net(name="net2", external=False)
        net3 = ckt.new_net(name="net3", external=False)

        layout.add_shape(layer=active, net=net1, shape=r1)
        layout.add_shape(layer=active, net=net1, shape=mps234.parts[0])
        layout.add_shape(layer=active, net=None, shape=mps234.parts[1])
        layout.add_shape(layer=active, net=net2, shape=mps234.parts[2])
        layout.add_shape(layer=active, net=net2, shape=mps567.parts[0])
        layout.add_shape(layer=active, net=None, shape=mps567.parts[1])
        layout.add_shape(layer=active, net=net3, shape=mps567.parts[2])
        layout.add_shape(layer=active, net=net3, shape=r8)

        # print("test_layout_multishapepart:merge()")
        lay_m = _kl.merge(layouter.layout)
        # print("done")

        for sl in lay_m._sublayouts:
            self.assertIsInstance(sl, _laylay._MaskShapesSubLayout)
            self.assertEqual(len(sl.shapes), 1)
            ms = sl.shapes[0]
            self.assertEqual(ms.mask, active.mask)
            if sl.net is None:
                self.assertIsInstance(ms.shape, _geo.MultiShape)
                self.assertTrue(all(
                    s.partshape in (r3, r6) for s in ms.shape.shapes
                ))
                self.assertTrue(all(
                    s.multipartshape.fullshape == r1_8 for s in ms.shape.shapes
                ))
            else:
                self.assertIsInstance(ms.shape, _geo.MultiPartShape._Part)
                # print(ms.shape.multipartshape.fullshape)
                self.assertEqual(ms.shape.multipartshape.fullshape, r1_8)
                if sl.net == net1:
                    self.assertEqual(ms.shape.partshape, r12)
                elif sl.net == net2:
                    self.assertEqual(ms.shape.partshape, r45)
                elif sl.net == net3:
                    self.assertEqual(ms.shape.partshape, r78)
                else:
                    assert False

    def test_library(self):
        from ...dummy import dummy_lib

        # Coverage of the library merge code
        _kl.merge(dummy_lib)



# SPDX-License-Identifier: AGPL-3.0-or-later OR GPL-2.0-or-later OR CERN-OHL-S-2.0+ OR Apache-2.0
from typing import cast

from pdkmaster.typing import GDSLayerSpecDict
from pdkmaster.technology import (
    property_ as _prp, geometry as _geo, net as _net, primitive as _prm, technology_ as _tch
)
from pdkmaster.design import (
    circuit as _ckt, layout as _lay,  library as _lbry, factory as _fab,
)
from pdkmaster.io.spice import SpicePrimsParamSpec

class MyNet(_net._Net):
    def __init__(self, *, name:str):
        super().__init__(name=name)


class EmptyTech(_tch.Technology):
    @property
    def name(self) -> str:
        return "Empty"
    @property
    def grid(self) -> float:
        return 0.005

    def __init__(self):
        super().__init__(
            primitives=_prm.Primitives(_prm.Base(type_=_prm.undopedBase)),
        )
empty_tech = EmptyTech()


class DummyTech(_tch.Technology):
    @property
    def name(self) -> str:
        return "Dummy"
    @property
    def grid(self) -> float:
        return 0.005
    
    def __init__(self):
        prims = _prm.Primitives(_prm.Base(type_=_prm.nBase))

        substrate = _prm.SubstrateMarker(name="substrate")
        prims += substrate

        nwell = _prm.Well(type_=_prm.nImpl, name="nwell", min_width=1.5, min_space=1.5)
        deepwell = _prm.DeepWell(
            name="deepwell", min_width=3.0, min_space=3.0,
            well=nwell, min_well_overlap=0.8, min_well_enclosure=0.8,
        )
        pwell = _prm.Well(type_=_prm.pImpl, name="pwell", min_width=1.5, min_space=1.5)
        nplus = _prm.Implant(type_=_prm.nImpl, name="nplus", min_width=1.0, min_space=1.0)
        pplus = _prm.Implant(type_=_prm.pImpl, name="pplus", min_width=1.0, min_space=1.0)
        hvox = _prm.Insulator(name="hvox", min_width=0.5, min_space=0.5)
        active = _prm.WaferWire(
            name="active", min_width=0.3, min_space=0.2,
            allow_in_substrate=True,
            implant=(nplus, pplus), min_implant_enclosure=_prp.Enclosure(0.2),
            implant_abut="none", allow_contactless_implant=False,
            well=(nwell, pwell), min_well_enclosure=_prp.Enclosure(1.0),
            min_substrate_enclosure=_prp.Enclosure(1.0),
            oxide=hvox, min_oxide_enclosure=_prp.Enclosure(0.2),
            allow_well_crossing=False,
        )
        poly = _prm.GateWire(name="poly", min_width=0.25, min_space=0.25)
        prims += (nwell, deepwell, pwell, nplus, pplus, hvox, active, poly)

        metalpin = _prm.Marker(name="metalpin")
        metal = _prm.MetalWire(
            name="metal",
            min_width=0.1, min_space=0.1, space_table=((0.2, 0.5), ((1.0, 1.0), 1.0)),
            min_area=0.05, min_density=0.20,
            pin=metalpin,
        )
        contact = _prm.Via(
            name="contact", width=0.35, min_space=0.35, bottom=(active, poly), top=metal,
            min_bottom_enclosure=_prp.Enclosure(0.2), min_top_enclosure=_prp.Enclosure(0.15),
        )
        prims += (contact, metalpin, metal)

        mimtop = _prm.MIMTop(
            name="MIMTop", min_width=0.2, min_space=0.2,
        )
        prims += mimtop

        metal2pin = _prm.Marker(name="metal2pin")
        metal2block = _prm.Marker(name="metal2block")
        metal2 = _prm.TopMetalWire(
            name="metal2", min_width=0.1, min_space=0.1,
            pin=metal2pin, blockage=metal2block,
        )
        metal2mark = _prm.Marker(name="metal2mark")
        metal2res = _prm.Resistor(
            name="metal2res", wire=metal2, min_length=0.5,
            indicator=metal2mark, min_indicator_extension=0.4,
            contact=None,
        )
        via = _prm.Via(
            name="via", width=0.35, min_space=0.35,
            bottom=(metal, mimtop), top=(metal2, metal2res),
            min_bottom_enclosure=_prp.Enclosure(0.2), min_top_enclosure=_prp.Enclosure(0.15),
        )
        prims += (via, metal2pin, metal2block, metal2, metal2mark, metal2res)

        silblock = _prm.ExtraProcess(
            name="silblock", min_width=0.4, min_space=0.4, grid=0.010,
        )
        resistor = _prm.Resistor(
            name="resistor", wire=active,
            implant=nplus, min_implant_enclosure=active.min_implant_enclosure[0],
            contact=contact, min_contact_space=0.2,
            indicator=silblock, min_indicator_extension=0.4,
        )
        polyres = _prm.Resistor(
            name="polyres", wire=poly, implant=pplus, contact=contact,
            min_implant_enclosure=_prp.Enclosure(0.3), min_contact_space=0.2,
            indicator=silblock, min_indicator_extension=0.4,
        )
        prims += (silblock, resistor, polyres)

        diodemark = _prm.Marker(name="diodemark")
        ndiode = _prm.Diode(
            name="ndiode", wire=active,
            indicator=diodemark, min_indicator_enclosure=_prp.Enclosure(0.2),
            implant=nplus, min_implant_enclosure=_prp.Enclosure(0.2),
        )
        pdiode = _prm.Diode(
            name="pdiode", wire=active,
            indicator=diodemark, min_indicator_enclosure=_prp.Enclosure(0.2),
            implant=pplus, min_implant_enclosure=_prp.Enclosure(0.2),
            well=nwell, min_well_enclosure=_prp.Enclosure(2.0),
        )
        prims += (diodemark, ndiode, pdiode)

        esd = _prm.Marker(name="esd")
        mosgate = _prm.MOSFETGate(
            name="mosgate", active=active, min_sd_width=0.35,
            max_l=10.0, max_w=50.0,
            poly=poly, min_polyactive_extension=0.45,
            contact=contact, min_contactgate_space=0.15,
        )
        hvgate = _prm.MOSFETGate(
            name="hvgate", active=active, oxide=hvox,
            min_sd_width=0.35, min_gate_space=0.5,
            poly=poly, min_polyactive_extension=0.45,
            contact=contact, min_contactgate_space=0.25,
            min_l=0.5, min_w=0.5,
        )
        esdgate = _prm.MOSFETGate(
            name="esdgate", active=active,
            oxide=hvox, min_gateoxide_enclosure=_prp.Enclosure(0.4),
            inside=esd, min_gateinside_enclosure=_prp.Enclosure(0.4),
            min_sd_width=0.35,
            poly=poly, min_polyactive_extension=0.45,
            min_l=0.5, min_w=0.5,
        )
        nmos = _prm.MOSFET(
            name="nmos", gate=mosgate,
            implant=nplus, min_gateimplant_enclosure=_prp.Enclosure(0.25),
        )
        pmos = _prm.MOSFET(
            name="pmos", gate=mosgate,
            implant=pplus, min_gateimplant_enclosure=_prp.Enclosure(0.25),
            well=nwell,
        )
        hvnmos = _prm.MOSFET(
            name="hvnmos", gate=hvgate,
            implant=nplus, min_gateimplant_enclosure=_prp.Enclosure(0.4),
        )
        hvpmos = _prm.MOSFET(
            name="hvpmos", gate=hvgate,
            min_l=0.8, max_l=20.0, min_w=0.8, max_w=30.0,
            implant=pplus, min_gateimplant_enclosure=_prp.Enclosure(0.4),
            well=nwell,
        )
        esdnmos = _prm.MOSFET(
            name="esdnmos", gate=esdgate, min_gate_space=2.0,
            min_polyactive_extension=0.4,
            implant=nplus, min_gateimplant_enclosure=_prp.Enclosure((0.4, 0.6)),
            contact=contact, min_contactgate_space=0.75,
            min_sd_width=0.8,
        )
        prims += (mosgate, hvgate, esdgate, nmos, pmos, esd, hvnmos, hvpmos, esdnmos)

        bip = _prm.Marker(name="bip")
        npn = _prm.Bipolar(name="npn", type_=_prm.npnBipolar, indicator=bip)
        pnp = _prm.Bipolar(
            name="pnp", type_=_prm.pnpBipolar, indicator=bip,
        )
        prims += (bip, npn, pnp)

        mimcap = _prm.MIMCapacitor(
            name="MIMCap", bottom=metal, top=mimtop, via=via,
            min_bottom_top_enclosure=_prp.Enclosure(0.2),
            min_bottomvia_top_space=0.25,
            min_top_via_enclosure=_prp.Enclosure(0.25),
            min_bottom_space=None, min_top2bottom_space=None,
        )
        mimcap2 = _prm.MIMCapacitor(
            name="MIMCap2", bottom=metal, top=mimtop, via=via,
            min_bottom_top_enclosure=_prp.Enclosure(0.2),
            min_bottomvia_top_space=0.25,
            min_top_via_enclosure=_prp.Enclosure(0.25),
            min_bottom_space=None, min_top2bottom_space=None,
        )
        prims += (mimcap, mimcap2)

        aux = _prm.Auxiliary(name="anything_goes")
        prims += aux

        prims += (
            _prm.Spacing(
                primitives1=(nplus, pplus), primitives2=mosgate, min_space=0.25,
            ),
            _prm.Spacing(
                primitives1=active, primitives2=hvox, min_space=0.2,
            ),
            _prm.Spacing(
                primitives1=(nwell, pplus), min_space=2.0,
            ),
            _prm.MinWidth(prim=active.in_(hvox), min_width=0.5),
            _prm.Enclosure(prim=active, by=esd, min_enclosure=_prp.Enclosure(0.1)),
            _prm.NoOverlap(prim1=nplus, prim2=pplus),
        )

        super().__init__(primitives=prims)
dummy_tech = DummyTech()
dummy_cktfab = _ckt.CircuitFactory(tech=dummy_tech)
dummy_layoutfab = _lay.LayoutFactory(tech=dummy_tech)

dummy_gdslayers: GDSLayerSpecDict = {
    mask.name: (i + 1) if i%2 == 0 else (i + 1, 1)
    for i, mask in enumerate(dummy_tech.designmasks)
}
dummy_gdslayers["anything_goes"] = None
dummy_textgdslayers: GDSLayerSpecDict = {
    "metalpin": (10, 10),
}

def _gen_spiceparams() -> SpicePrimsParamSpec:
    prims = dummy_tech.primitives

    params = SpicePrimsParamSpec()
    params.add_device_params(
        prim=cast(_prm.Resistor, prims.metal2res),
        model="mres2", sheetres=10.0,
    )
    params.add_device_params(
        prim=cast(_prm.Resistor, prims.resistor), is_subcircuit=True,
    )
    params.add_device_params(
        prim=cast(_prm.Resistor, prims.polyres), is_subcircuit=True,
        model="polyres", subcircuit_paramalias={"width": "w", "length": "l"},
    )

    params.add_device_params(prim=cast(_prm.Diode, prims.ndiode))
    params.add_device_params(prim=cast(_prm.Diode, prims.pdiode), is_subcircuit=True)

    params.add_device_params(prim=cast(_prm.MOSFET, prims.nmos))
    params.add_device_params(prim=cast(_prm.MOSFET, prims.pmos), is_subcircuit=True)
    params.add_device_params(prim=cast(_prm.MOSFET, prims.hvnmos))
    params.add_device_params(prim=cast(_prm.MOSFET, prims.hvpmos))
    params.add_device_params(prim=cast(_prm.MOSFET, prims.esdnmos), model="nmos_5v_esd")

    params.add_device_params(prim=cast(_prm.Bipolar, prims.npn))
    params.add_device_params(prim=cast(_prm.Bipolar, prims.pnp), is_subcircuit=True)

    params.add_device_params(prim=cast(_prm.MIMCapacitor, prims.MIMCap))
    params.add_device_params(
        prim=cast(_prm.MIMCapacitor, prims.MIMCap2),
        model="cap2", subcircuit_paramalias={"width": "w", "height": "h"},
    )

    return params
dummy_prims_spiceparams = _gen_spiceparams()

dummy_lib = _lbry.Library(name="dummy_lib", tech=dummy_tech)
dummy_fab = _fab.BaseCellFactory(lib=dummy_lib, cktfab=dummy_cktfab, layoutfab=dummy_layoutfab)
def _lib_init():
    # The dummy_lib will be initialized with different shapes to maximize
    # code coverage when being used for example in export code etc.
    prims = dummy_tech.primitives
    metal = cast(_prm.TopMetalWire, prims.metal)

    rect1 = _geo.Rect(left=0.0, bottom=0.0, right=1.0, top=1.0)
    rect2 = _geo.Rect(left=1.0, bottom=0.0, right=2.0, top=1.0)
    rect12 = _geo.Rect(left=0.0, bottom=0.0, right=2.0, top=1.0)
    lshape = _geo.Polygon.from_floats(points=(
        (0.0, 0.0), (0.0, 3.0), (1.0, 3.0), (1.0, 1.0), (2.0, 1.0), (2.0, 0.0), (0.0, 0.0),
    ))

    # cell1: rect shape
    cell1 = dummy_fab.new_cell(name="cell1")
    ckt = cell1.new_circuit()
    layouter = cell1.new_circuitlayouter()
    layout = layouter.layout

    i = ckt.new_net(name="i", external=True)
    layouter.add_wire(net=i, wire=metal, pin=prims.metalpin, shape=rect1)
    layouter.layout.boundary = rect1

    # cell2: cell instance
    cell2 = dummy_fab.new_cell(name="cell2")
    ckt = cell2.new_circuit()
    layouter = cell2.new_circuitlayouter()

    inst = ckt.instantiate(cell1, name="inst")
    i = ckt.new_net(name="i", external=True, childports=inst.ports.i)

    layouter.place(object_=inst, origin=_geo.Point(x=1.0, y=0.0))
    layouter.layout.boundary = rect1

    # cell3: polygon
    cell3 = dummy_fab.new_cell(name="cell3")
    ckt = cell3.new_circuit()
    layouter = cell3.new_circuitlayouter()

    i = ckt.new_net(name="i", external=True)
    layouter.add_wire(net=i, wire=metal, pin=prims.metalpin, shape=lshape)
    layouter.layout.boundary = lshape.bounds

    # cell4: multipartshape
    cell4 = dummy_fab.new_cell(name="cell4")
    ckt = cell4.new_circuit()
    layouter = cell4.new_circuitlayouter()
    layout = layouter.layout

    mps = _geo.MultiPartShape(fullshape=rect12, parts=(rect1, rect2))
    
    i1 = ckt.new_net(name="i1", external=False)
    i2 = ckt.new_net(name="i2", external=False)

    layout.add_shape(layer=metal, net=i1, shape=mps.parts[0])
    layout.add_shape(layer=metal, net=i2, shape=mps.parts[1])

    # cell5: primitive instance
    cell5 = dummy_fab.new_cell(name="cell5")
    ckt = cell5.new_circuit()
    layouter = cell5.new_circuitlayouter()
    layout = layouter.layout

    res = ckt.instantiate(
        cast(_prm.Resistor, prims.resistor), name="res",
        width=1.0, height=5.0,
    )

    ckt.new_net(name="p1", external=True, childports=res.ports.port1)
    ckt.new_net(name="p2", external=True, childports=res.ports.port2)

    layouter.place(res)
_lib_init()



# SPDX-License-Identifier: AGPL-3.0-or-later OR GPL-2.0-or-later OR CERN-OHL-S-2.0+ OR Apache-2.0
# type: ignore
from textwrap import dedent
import unittest

from pdkmaster import _util
from pdkmaster.technology import (
    property_ as _prp, mask as _msk, edge as _edg, geometry as _geo
)
from pdkmaster.io import klayout as _kl

from ...dummy import (
    dummy_tech, dummy_gdslayers, dummy_textgdslayers, dummy_prims_spiceparams, dummy_lib,
    empty_tech,
)

import pya


class FileExportTest(unittest.TestCase):
    def test_maskconverter(self):
        from pdkmaster.technology.wafer_ import wafer
        from pdkmaster.io.klayout.export import _MaskConverter

        mc = _MaskConverter(tech=dummy_tech)

        mask = _msk.DesignMask(name="layer0")
        alias = mask.alias("0layer:s1.s2")
        mask2 = _msk.DesignMask(name="layer1")

        self.assertEqual(mc(mask), "layer0")
        self.assertEqual(mc(alias), "_0layer__s1_s2")
        self.assertEqual(mc(_msk.Join((mask, mask2))), "(layer0+layer1)")
        self.assertEqual(mc(_msk.Intersect((mask, mask2))), "(layer0&layer1)")
        self.assertEqual(mc(mask.remove(mask2)), "(layer0-layer1)")
        self.assertEqual(mc(wafer), "extent.sized(1.0)")

    def test_waferempty(self):
        from pdkmaster.technology.wafer_ import wafer
        from pdkmaster.io.klayout.export import _MaskConverter

        mc = _MaskConverter(tech=empty_tech)
        self.assertEqual(mc(wafer), "extent")

    def test_edgeconverter(self):
        from pdkmaster.io.klayout import export
        # _edge_conv needs _mask_conv to be initialized
        export._mask_conv = export._MaskConverter(tech=dummy_tech)

        ec = export._edge_conv

        mask = _msk.DesignMask(name="mask")
        edge = _edg.MaskEdge(mask)
        mask2 = _msk.DesignMask(name="mask2")
        edge2 = _edg.MaskEdge(mask2)

        interact1 = edge.interact_with(mask2)
        interact2 = edge.interact_with(edge2)

        self.assertEqual(ec(edge), "mask.edges")
        self.assertEqual(ec(interact1), "mask.edges.interacting(mask2)")
        self.assertEqual(ec(interact2), "mask.edges.interacting(mask2.edges)")
        self.assertEqual(ec(_edg.Join((edge, edge2))), "(mask.edges+mask2.edges)")
        self.assertEqual(ec(_edg.Join((edge, mask2))), "(mask.edges+mask2)")
        self.assertEqual(ec(_edg.Intersect((edge, edge2))), "(mask.edges&mask2.edges)")
        self.assertEqual(ec(_edg.Intersect((edge, mask2))), "(mask.edges&mask2)")

        export._mask_conv = None

    def test_stralias(self):
        from pdkmaster.io.klayout import export
        export._mask_conv = export._MaskConverter(tech=dummy_tech)

        mask = _msk.DesignMask(name="mask")
        alias = mask.alias("alias")
        self.assertEqual(export._str_alias(alias), "alias = mask\n")

        export._mask_conv = None

    def test_ruleconverter(self):
        from pdkmaster.io.klayout import export
        # _edge_conv needs _mask_conv to be initialized
        export._mask_conv = export._MaskConverter(tech=dummy_tech)

        rc = export._rule_conv

        mask = _msk.DesignMask(name="mask")
        alias = mask.alias("alias")
        edge = _edg.MaskEdge(mask)
        mask2 = _msk.DesignMask(name="mask2")

        self.assertEqual(
            rc(mask.area >= 0.1),
            dedent("""
                # mask.area >= 0.1
                mask.with_area(nil, 0.1).output(
                    "mask area", "mask minimum area: 0.1µm"
                )
            """[1:]),
        )
        self.assertEqual(
            rc(mask.density >= 0.25),
            dedent("""
                # mask.density >= 0.25
                mask_mindens = polygon_layer
                dens_check(mask_mindens, mask, 0.25, 1)
                mask_mindens.output(
                    "mask density", "mask minimum density: 25%"
                )
            """[1:]),
        )
        self.assertEqual(
            rc(mask.density <= 0.80, allow_unimplented=True),
            dedent("""
                # mask.density <= 0.8
                # Not supported
            """[1:]),
        )
        self.assertEqual(
            rc(alias.space >= 1.0, allow_unimplented=True),
            dedent("""
                # alias.space >= 1.0
                alias.space(1.0).output(
                    "alias space", "alias minimum space: 1.0µm"
                )
            """[1:]),
        )
        self.assertEqual(
            rc(edge.length >= 0.5),
            dedent("""
                # edge(mask).length >= 0.5
                mask.edges.with_length(nil, 0.5).output(
                    "mask.edges length",
                    "Minimum length of mask.edges: 0.5µm"
                )
            """[1:]),
        )
        self.assertEqual(
            rc(_msk.Spacing(mask, mask2) >= 1.0),
            dedent("""
                # space(mask,mask2) >= 1.0
                mask.separation(mask2, 1.0, square).output(
                    "mask:mask2 spacing",
                    "Minimum spacing between mask and mask2: 1.0µm"
                )
            """[1:]),
        )
        self.assertEqual(
            rc(mask.extend_over(mask2) >= 0.1),
            dedent("""
                # mask.extend_over(mask2) >= 0.1
                extend_check(mask2, mask, 0.1).output(
                    "mask:mask2 extension",
                    "Minimum extension of mask of mask2: 0.1µm"
                )
            """[1:]),
        )
        self.assertEqual(
            rc(mask.enclosed_by(mask2) >= 0.1),
            dedent("""
                # mask.enclosed_by(mask2) >= Enclosure(0.1)
                mask2.enclosing(mask, 0.1).output(
                    "mask2:mask enclosure",
                    "Minimum enclosure of mask2 around mask: 0.1µm"
                )
            """[1:]),
        )
        self.assertEqual(
            rc(edge.enclosed_by(mask2) >= 0.1),
            dedent("""
                # edge(mask).enclosed_by(mask2) >= 0.1
                mask2.edges.enclosing(mask.edges, 0.1).output(
                    "mask2.edges:mask.edges enclosure",
                    "Minimum enclosure of mask2.edges around mask.edges: 0.1µm"
                )
            """[1:]),
        )
        self.assertEqual(
            rc(mask.enclosed_by(mask2) >= _prp.Enclosure((0.2, 0.1))),
            dedent("""
                # mask.enclosed_by(mask2) >= Enclosure((0.2,0.1))
                oppenc_check(mask, mask2, 0.1, 0.2).output(
                    "mask2:mask asymmetric enclosure",
                    "Minimum enclosure of mask2 around mask: 0.1µm minimum, 0.2µm opposite"
                )
            """[1:]),
        )
        self.assertEqual(
            rc(_msk.Spacing(mask.parts_with(mask.width >= 3.0), mask) >= 0.8),
            dedent("""
                # space(mask.parts_with(mask.width >= 3.0),mask) >= 0.8
                space4width_check(mask, 3.0, 0.8).output(
                    "mask table spacing",
                    "Minimum mask spacing for 3.0µm width: 0.8µm"
                )
            """[1:]),
        )
        self.assertEqual(
            rc(_msk.Spacing(mask, mask.parts_with(mask.width >= 3.0)) >= 0.8),
            dedent("""
                # space(mask,mask.parts_with(mask.width >= 3.0)) >= 0.8
                space4width_check(mask, 3.0, 0.8).output(
                    "mask table spacing",
                    "Minimum mask spacing for 3.0µm width: 0.8µm"
                )
            """[1:]),
        )
        self.assertEqual(
            rc(mask.width == 0.5),
            dedent("""
                # mask.width == 0.5
                width_check(mask, 0.5).output(
                    "mask width", "mask width: 0.5µm"
                )
            """[1:]),
        )
        self.assertEqual(
            rc(mask.area == 0.0),
            dedent("""
                # mask.area == 0.0
                mask.output("mask empty")
            """[1:]),
        )
        with self.assertRaises(ValueError):
            rc(mask.area == 1.0)
        self.assertEqual(
            rc(edge.length == 0.0),
            dedent("""
                # edge(mask).length == 0.0
                mask.edges.output("mask.edges empty")
            """[1:]),
        )
        with self.assertRaises(ValueError):
            rc(edge.length == 0.1)
        self.assertEqual(
            rc(alias),
            dedent("""
                # mask.alias(alias)
                alias = mask
            """[1:]),
        )
        self.assertEqual(
            rc(_msk.Connect(mask, mask2)),
            dedent("""
                # connect(mask,mask2)
                connect(mask, mask2)
            """[1:]),
        )

        export._mask_conv = None

    def test_shapeexporter(self):
        from pdkmaster.io.klayout.export import _ShapeExporter

        se = _ShapeExporter(export_fullshape=False)

        rect1 = _geo.Rect(left=0.0, bottom=0.0, right=1.0, top=1.0)
        rect2 = _geo.Rect(left=1.0, bottom=0.0, right=2.0, top=1.0)
        rect12 = _geo.Rect(left=0.0, bottom=0.0, right=2.0, top=1.0)
        shapes = (rect1, rect2)
        rs = _geo.RepeatedShape(
            shape=rect1, n=2, n_dxy=_geo.Point(x=0.0, y=2.0), offset0=_geo.origin,
        )
        ms1 = _geo.MultiShape(shapes=shapes)
        ms2 = _geo.MultiShape(shapes=(*shapes, rs))
        mps = _geo.MultiPartShape(fullshape=rect12, parts=shapes)
        part0 = mps.parts[0]
        part1 = mps.parts[1]
        rmps = _geo.RepeatedShape(
            shape=mps, n=3, n_dxy=_geo.Point(x=0.0, y=10.0), offset0=_geo.origin,
        )

        rr = _geo.RectRing(outer_bound=rect1, rect_width=0.3, min_rect_space=0.3)

        self.assertEqual(
            se(_geo.Point(x=0.0, y=0.0)),
            pya.DPoint(0.0, 0.0),
        )
        self.assertEqual(
            se(_geo.Line(point1=_geo.Point(x=0.0, y=0.0), point2=_geo.Point(x=0.0, y=1.0))),
            pya.DPath((pya.DPoint(0.0, 0.0), pya.DPoint(0.0, 1.0)), 0.0),
        )
        points = (
            (0.0, 0.0), (0.0, 2.0), (2.0, 2.0), (2.0, 1.0), (1.0, 1.0), (1.0, 0.0), (0.0, 0.0),
        )
        self.assertEqual(
            se(_geo.Polygon.from_floats(points=points)),
            pya.DSimplePolygon(tuple(pya.DPoint(x, y) for x, y in points[:-1])),
        )
        self.assertEqual(se(rect1), pya.DBox(0.0, 0.0, 1.0, 1.0))
        self.assertEqual(
            se(_geo.Label(origin=_geo.origin, text="label")),
            pya.DText("label", 0.0, 0.0),
        )
        self.assertEqual(set(se(ms1)), set(se(shape) for shape in shapes))
        self.assertEqual(set(se(ms2)), set(se(shape) for shape in ms2.pointsshapes))

        self.assertEqual(
            set(se(rmps)),
            set(se(ps) for ps in rmps.pointsshapes),
        )

        self.assertEqual(se(mps), se(mps.fullshape))
        self.assertEqual(se(part0), se(part0.partshape))

        se2 = _ShapeExporter(export_fullshape=True)
        self.assertEqual(se2(part0), se2(mps.fullshape))
        self.assertIsNone(se2(part1))

        self.assertEqual(set(se(rr)), set(se(shape) for shape in rr.pointsshapes))

    def test_tech(self):
        # We just generate the code to get the code coverage.
        # Correctness of out is assumed to be checked with lower level unit tests
        # on the Exporter classes and higher level test using gds files etc.
        _kl.FileExporter(
            tech=dummy_tech, gds_layers=dummy_gdslayers,
            prims_spiceparams=dummy_prims_spiceparams,
        )()


class Export2DBTest(unittest.TestCase):
    # We don't foresee special unit test for _LayoutExporter class
    # All test of that class to be done through export2db unit tests
    rect_geo = _geo.Rect(left=0.0, bottom=0.0, right=1.0, top=1.0)
    rect_kl = pya.DBox(0.0, 0.0, 1.0, 1.0)

    def test_error(self):
        with self.assertRaises(TypeError):
            _kl.export2db(3, gds_layers=dummy_gdslayers)

    def test_rect(self):
        rect_conv = _kl.export2db(self.rect_geo)
        self.assertEqual(rect_conv, self.rect_kl)

        with self.assertRaises(AssertionError):
            _kl.export2db(self.rect_geo, gds_layers=dummy_gdslayers)

    def test_maskshapes(self):
        # This mainly runs code to get coverage but does not extensive test of correctness
        # of results. This is assumed to be done with lower level unit tests.
        mask = _util.get_first_of(dummy_tech.designmasks)

        rect1 = _geo.Rect(left=0.0, bottom=0.0, right=1.0, top=1.0)
        rect2 = _geo.Rect(left=2.0, bottom=0.0, right=3.0, top=1.0)

        ms = _geo.MultiShape(shapes=(rect1, rect2))

        msr1 = _geo.MaskShape(mask=mask, shape=rect1)
        msr2 = _geo.MaskShape(mask=mask, shape=ms)

        with self.assertRaises(AssertionError):
            _kl.export2db(msr1)
        kllay = _kl.export2db(msr1, gds_layers=dummy_gdslayers)
        self.assertIsInstance(kllay, pya.Layout)
        kllay = _kl.export2db(msr2, gds_layers=dummy_gdslayers)
        self.assertIsInstance(kllay, pya.Layout)

    def test_cells_library(self):
        # This mainly runs code to get coverage but does not extensive test of correctness
        # of results. This is assumed to be done with lower level unit tests.
        _kl.export2db(
            dummy_lib.cells.cell1, gds_layers=dummy_gdslayers, add_pin_label=True, merge=True,
        )

        with self.assertRaises(AssertionError):
            _kl.export2db(dummy_lib, gds_layers=dummy_gdslayers, cell_name="test")
        _kl.export2db(dummy_lib, gds_layers=dummy_gdslayers, add_pin_label=True, merge=True)
        _kl.export2db(
            dummy_lib, gds_layers=dummy_gdslayers, textgds_layers={},
            add_pin_label=True, merge=True,
        )
        _kl.export2db(
            dummy_lib, gds_layers=dummy_gdslayers, textgds_layers=dummy_textgdslayers,
            add_pin_label=True, merge=True,
        )
